//                              -*- Mode: C++ -*- 
// 
// uC++ Version 5.3.0, Copyright (C) Peter A. Buhr 1999
// 
// ClientINET2.cc -- Client for INET/datagram socket test. Client reads from
// standard input, writes the data to the server, reads the data from the
// server, and writes that data to standard output.
// 
// Author           : Peter A. Buhr
// Created On       : Thu Apr 29 16:04:50 1999
// Last Modified By : Peter A. Buhr
// Last Modified On : Wed Nov 30 13:20:52 2005
// Update Count     : 50
// 

#include <uC++.h>
#include <uSemaphore.h>
#include <uSocket.h>
#include <iostream>
using std::cin;
using std::cout;
using std::cerr;
using std::osacquire;
using std::endl;

#define EOD '\377'
// minimum buffer size is 2, 1 character and string terminator, '\0'
#define BufferSize (65)

int rcnt = 0, wcnt = 0;

// Datagram sockets are lossy (i.e., drop packets). To prevent clients from
// flooding the server with packets, resulting in dropped packets, a semaphore
// is used to synchronize the reader and writer tasks so at most N writes occur
// before a read. As well, if the buffer size is increase substantially, it may
// be necessary to decrease N to ensure the server buffer does not fill.

const int MaxWriteBeforeRead = 5;
uSemaphore readSync(MaxWriteBeforeRead);

_Task reader {
	uSocketClient &client;

	void main() {
		char buf[BufferSize];
		int len;

		for ( ;; ) {
			len = client.recvfrom( buf, sizeof(buf) );
			readSync.V();
			// osacquire( cerr ) << "Client::reader read len:" << len << endl;
		  if ( len == 0 ) uAbort( "(uSocketClient &)0x%p : EOF ecountered without EOD", &client );
			// The EOD character can be piggy-backed onto the end of the message.
		  if ( buf[len - 1] == EOD ) {
			  rcnt += len - 1;
			  cout.write( buf, len - 1 );				// do not write the EOD
			  break; }
			rcnt += len;
			cout.write( buf, len );
		} // for
	} // reader::main
  public:
	reader( uSocketClient &client ) : client ( client ) {
	} // reader::reader
}; // reader

_Task writer {
	uSocketClient &client;

	void main() {
		char buf[BufferSize];

		for ( ;; ) {
			cin.get( buf, sizeof(buf), '\0' );			// leave room for string terminator
			int len = strlen( buf );
			// osacquire( cerr ) << "Client::writer read len:" << len << endl;
		  if ( buf[0] == '\0' ) break;
			wcnt += len;
			readSync.P();
			client.sendto( buf, len );
		} // for
		buf[0] = EOD;
		readSync.P();
		client.sendto( buf, sizeof(char) );
	} // writer::main
  public:
	writer( uSocketClient &client ) : client( client ) {
	} // writer::writer
}; // writer

void uMain::main() {
	switch ( argc ) {
	  case 2:
		break;
	  default:
		cerr << "Usage: " << argv[0] << " port-number" << endl;
		uExit( -1 );
	} // switch

	uSocketClient client( atoi( argv[1] ), SOCK_DGRAM ); // connection to server
	{
		reader rd( client );							// emit worker to read from server and write to output
		writer wr( client );							// emit worker to read from input and write to server
	}
	if ( wcnt != rcnt ) {
		uAbort( "not all data transfered\n" );
	} // if
} // uMain::main

// Local Variables: //
// tab-width: 4 //
// compile-command: "u++-work -o Client ClientINET2.cc" //
// End: //
