//                              -*- Mode: C++ -*- 
// 
// uC++ Version 5.3.0, Copyright (C) Peter A. Buhr 1994
// 
// uFile.cc -- 
// 
// Author           : Peter Buhr
// Created On       : Tue Mar 29 16:42:36 1994
// Last Modified By : Peter A. Buhr
// Last Modified On : Fri Jun 24 14:48:32 2005
// Update Count     : 293
//
// This  library is free  software; you  can redistribute  it and/or  modify it
// under the terms of the GNU Lesser General Public License as published by the
// Free Software  Foundation; either  version 2.1 of  the License, or  (at your
// option) any later version.
// 
// This library is distributed in the  hope that it will be useful, but WITHOUT
// ANY  WARRANTY;  without even  the  implied  warranty  of MERCHANTABILITY  or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License
// for more details.
// 
// You should  have received a  copy of the  GNU Lesser General  Public License
// along  with this library.
// 


#define __U_KERNEL__
#include <uC++.h>
#include <uFile.h>

//#include <uDebug.h>

#include <cstring>					// strerror
#include <unistd.h>					// read, write, close, etc.
#include <sys/uio.h>					// readv, writev


//######################### uFile #########################


void uFile::access() {
    accessCnt += 1;
} // uFile::access


void uFile::unaccess() {
    accessCnt -= 1;
} // uFile::unaccess


void uFile::uCreateFile( const char *name ) {
    uName = new char[strlen( name ) + 1];
    strcpy( uName, name );
    accessCnt = 0;
} // uFile::uCreateFile


uFile::uFile( const char *name ) {
    uCreateFile( name );
} // uFile::uFile


uFile::~uFile() {
    TerminateFailure temp( *this, accessCnt, "terminating access with outstanding accessor(s)" );
    if ( uName != 0 ) delete uName;			// safe to delete name as the name is copied in the exception object
    if ( accessCnt != 0 ) {
	if ( ! std::uncaught_exception() ) _Throw temp;
    } // if
} // uFile::~uFile


const char *uFile::getName() const {
    return
#ifdef __U_DEBUG__
	( uName == NULL || uName == (const char *)-1 ) ? "*unknown*" : // storage might be scrubbed
#endif // __U_DEBUG__
	uName;
} // uFile::getName


void uFile::status( struct stat &buf ) {
    int code;

    for ( ;; ) {
	code = ::stat( uName, &buf );
      if ( code != -1 ) break;
      if ( errno != EINTR ) break;			// timer interrupt ?
    } // for
    if ( code == -1 ) {
	_Resume uFile::StatusFailure( *this, buf, "could not obtain statistical information for file" );
    } // if
} // uFile::status


//######################### uFileIO #########################


uFileIO::uFileIO( uIOaccess &acc ) : access( acc ) {
} // uFileIO::uFileIO


uFileIO::~uFileIO() {
} // uFileIO::~uFileIO


int uFileIO::_read( char *buf, int len ) {
    if ( access.poll.uGetStatus() == uPoll::uPollOnDemand ) access.poll.uSetPollFlag( access.fd );
    int rlen = ::read( access.fd, buf, len );
    int terrno = errno;					// preserve errno across calls
#ifdef __U_DEBUG_H__
    uDebugPrt( "(uFileIO &)0x%p._read attempting read to fd:%d, rlen:%d, errno:%d\n", this, access.fd, rlen, errno );
#endif // __U_DEBUG_H__
    if ( access.poll.uGetStatus() == uPoll::uPollOnDemand ) access.poll.uClearPollFlag( access.fd );
    errno = terrno;
    return rlen;
} // uFileIO::_read


int uFileIO::_readv( const struct iovec *iov, int iovcnt ) {
    if ( access.poll.uGetStatus() == uPoll::uPollOnDemand ) access.poll.uSetPollFlag( access.fd );
    int rlen = ::readv( access.fd, iov, iovcnt );
    int terrno = errno;					// preserve errno across calls
#ifdef __U_DEBUG_H__
    uDebugPrt( "(uFileIO &)0x%p._readv attempting readv to fd:%d, rlen:%d, errno:%d\n", this, access.fd, rlen, errno );
#endif // __U_DEBUG_H__
    if ( access.poll.uGetStatus() == uPoll::uPollOnDemand ) access.poll.uClearPollFlag( access.fd );
    errno = terrno;
    return rlen;
} // uFileIO::_readv


int uFileIO::_write( char *buf, int len ) {
    if ( access.poll.uGetStatus() == uPoll::uPollOnDemand ) access.poll.uSetPollFlag( access.fd );
    int wlen = ::write( access.fd, buf, len );
    int terrno = errno;					// preserve errno across calls
#ifdef __U_DEBUG_H__
    uDebugPrt( "(uFileIO &)0x%p._write attempting write to fd:%d, wlen:%d, errno:%d\n", this, access.fd, wlen, errno );
#endif // __U_DEBUG_H__
    if ( access.poll.uGetStatus() == uPoll::uPollOnDemand ) access.poll.uClearPollFlag( access.fd );
    errno = terrno;
    return wlen;
} // uFileIO::_write


int uFileIO::_writev( const struct iovec *iov, int iovcnt ) {
    if ( access.poll.uGetStatus() == uPoll::uPollOnDemand ) access.poll.uSetPollFlag( access.fd );
    int wlen = ::writev( access.fd, iov, iovcnt );
    int terrno = errno;					// preserve errno across calls
#ifdef __U_DEBUG_H__
    uDebugPrt( "(uFileIO &)0x%p._writev attempting writev to fd:%d, wlen:%d, errno:%d\n", this, access.fd, wlen, errno );
#endif // __U_DEBUG_H__
    if ( access.poll.uGetStatus() == uPoll::uPollOnDemand ) access.poll.uClearPollFlag( access.fd );
    errno = terrno;
    return wlen;
} // uFileIO::_writev


int uFileIO::read( char *buf, int len, uDuration *timeout ) {
    int rlen;

#ifdef __U_DEBUG_H__
    uDebugPrt( "(uFileIO &)0x%p.read attempting read to fd:%d\n", this, access.fd );
#endif // __U_DEBUG_H__

    for ( ;; ) {
	rlen = _read( buf, len );
      if ( rlen != -1 ) break;
      if ( ! ( errno == EINTR || errno == U_EWOULDBLOCK ) ) break;
	if ( errno == U_EWOULDBLOCK ) {
	    int mask = uCluster::uReadSelect;
	    if ( timeout == NULL ) {
		uThisCluster().uSelect( access.fd, mask );
	    } else {
		timeval t = *timeout;			// convert to timeval for select
		if ( uThisCluster().uSelect( access.fd, mask, &t ) == 0 ) { // timeout ?
		    readTimeout( buf, len, timeout, "read" );
		} // if
	    } // if
	} // if
    } // for    

    if ( rlen == -1 ) {
	readFailure( buf, len, timeout, "read" );
    } // if

#ifdef __U_DEBUG_H__
    uDebugPrt( "(uFileIO &)0x%p.read completed read to fd:%d, read:%d bytes\n", this, access.fd, rlen );
#endif // __U_DEBUG_H__
    return rlen;
} // uFileIO::read


int uFileIO::readv( const struct iovec *iov, int iovcnt, uDuration *timeout ) {
    int rlen;

#ifdef __U_DEBUG_H__
    uDebugPrt( "(uFileIO &)0x%p.readv attempting read to fd:%d\n", this, access.fd );
#endif // __U_DEBUG_H__

    for ( ;; ) {
	rlen = _readv( iov, iovcnt );
      if ( rlen != -1 ) break;
      if ( ! ( errno == EINTR || errno == U_EWOULDBLOCK ) ) break;
	if ( errno == U_EWOULDBLOCK ) {
	    int mask = uCluster::uReadSelect;
	    if ( timeout == NULL ) {
		uThisCluster().uSelect( access.fd, mask );
	    } else {
		timeval t = *timeout;			// convert to timeval for select
		if ( uThisCluster().uSelect( access.fd, mask, &t ) == 0 ) { // timeout ?
		    readTimeout( (const char *)iov, iovcnt, timeout, "readv" );
		} // if
	    } // if
	} // if
    } // for    
    if ( rlen == -1 ) {
	readFailure( (const char *)iov, iovcnt, timeout, "readv" );
    } // if

#ifdef __U_DEBUG_H__
    uDebugPrt( "(uFileIO &)0x%p.readv completed read to fd:%d, read:%d bytes\n", this, access.fd, rlen );
#endif // __U_DEBUG_H__
    return rlen;
} // uFileIO::readv


int uFileIO::write( char *buf, int len, uDuration *timeout ) {
    int count, wlen;

#ifdef __U_DEBUG_H__
    uDebugPrt( "(uFileIO &)0x%p.write attempting write to fd:%d\n", this, access.fd );
#endif // __U_DEBUG_H__

    for ( count = 0; count < len; count += wlen ) {	// ensure all data is written
	for ( ;; ) {
	    wlen = _write( buf + count, len - count );
	  if ( wlen != -1 ) break;
	  if ( ! ( errno == EINTR || errno == U_EWOULDBLOCK ) ) break;
	    if ( errno == U_EWOULDBLOCK ) {
		int mask = uCluster::uWriteSelect;
		if ( timeout == NULL ) {
		    uThisCluster().uSelect( access.fd, mask );
		} else {
		    timeval t = *timeout;		// convert to timeval for select
		    if ( uThisCluster().uSelect( access.fd, mask, &t ) == 0 ) { // timeout ?
			writeTimeout( buf, len, timeout, "write" );
		    } // if
		} // if
	    } // if
	} // for

	if ( wlen == -1 ) {
	    // EIO means the write is to stdout but the shell has terminated (I
	    // think). Normally, people want this to work as if stdout is
	    // magically redirected to /dev/null, instead of aborting the
	    // program.

      if ( errno == EIO ) break;

      	    writeFailure( buf, len, timeout, "write" );
	} // if
    } // for

#ifdef __U_DEBUG_H__
    uDebugPrt( "(uFileIO &)0x%p.write completed write to fd:%d, wrote:%d bytes\n", this, access.fd, len );
#endif // __U_DEBUG_H__
    return len;						// always return the specified length
} // uFileIO::write


int uFileIO::writev( const struct iovec *iov, int iovcnt, uDuration *timeout ) {
    int wlen;

#ifdef __U_DEBUG_H__
    uDebugPrt( "(uFileIO &)0x%p.writev attempting write to fd:%d\n", this, access.fd );
#endif // __U_DEBUG_H__

    for ( ;; ) {
	wlen = _writev( iov, iovcnt );
      if ( wlen != -1 ) break;
      if ( ! ( errno == EINTR || errno == U_EWOULDBLOCK ) ) break;
	if ( errno == U_EWOULDBLOCK ) {
	    int mask = uCluster::uWriteSelect;
	    if ( timeout == NULL ) {
		uThisCluster().uSelect( access.fd, mask );
	    } else {
		timeval t = *timeout;		// convert to timeval for select
		if ( uThisCluster().uSelect( access.fd, mask, &t ) == 0 ) { // timeout ?
		    writeTimeout( (const char *)iov, iovcnt, timeout, "writev" );
		} // if
	    } // if
	} // if
    } // for

    // EIO means the write is to stdout but the shell has terminated (I
    // think). Normally, people want this to work as if stdout is magically
    // redirected to /dev/null, instead of aborting the program.

    if ( wlen == -1 && errno != EIO ) {
	writeFailure( (const char *)iov, iovcnt, timeout, "writev" );
    } // if

#ifdef __U_DEBUG_H__
    uDebugPrt( "(uFileIO &)0x%p.writev completed write to fd:%d, wrote:%d bytes\n", this, access.fd, wlen );
#endif // __U_DEBUG_H__
    return wlen;
} // uFileIO::writev


int uFileIO::fd() {
    return access.fd;
} // uFileIO::fd


//######################### uFileAccess #########################


void uFileAccess::readFailure( const char *buf, const int len, const uDuration *timeout, const char *const op ) {
    char msg[32];
    strcat( strcat( strcpy( msg, "file " ), op ), " fails" );
    _Throw uFileAccess::ReadFailure( *this, buf, len, timeout, msg );
} // uFileAccess::readFailure


void uFileAccess::readTimeout( const char *buf, const int len, const uDuration *timeout, const char *const op ) {
    char msg[32];
    strcat( strcpy( msg, "timeout during file " ), op );
    _Throw uFileAccess::ReadTimeout( *this, buf, len, timeout, msg );
} // uFileAccess::readTimeout


void uFileAccess::writeFailure( const char *buf, const int len, const uDuration *timeout, const char *const op ) {
    char msg[32];
    strcat( strcat( strcpy( msg, "file " ), op ), " fails" );
    _Throw uFileAccess::WriteFailure( *this, buf, len, timeout, msg );
} // uFileAccess::writeFailure


void uFileAccess::writeTimeout( const char *buf, const int len, const uDuration *timeout, const char *const op ) {
    char msg[32];
    strcat( strcpy( msg, "timeout during file " ), op );
    _Throw uFileAccess::WriteTimeout( *this, buf, len, timeout, msg );
} // uFileAccess::writeTimeout


uFileAccess::uFileAccess( int fd, uFile &f ) : uFileIO( access ), ufile( f ) {
    access.fd = fd;
    access.poll.uSetStatus( uPoll::uPollOnDemand );
    ufile.access();
} // uFileAccess::uFileAccess


uFileAccess::uFileAccess( uFile &f, int flags, int mode ) : uFileIO( access ), ufile( f ) {
    for ( ;; ) {
	access.fd = ::open( ufile.uName, flags, mode );
      if ( access.fd != -1 ) break;
      if ( errno != EINTR ) break;			// timer interrupt ?
    } // for
    if ( access.fd == -1 ) {
        _Throw uFileAccess::OpenFailure( *this, flags, mode, "unable to access file" );
    } // if
    access.poll.uComputeStatus( access.fd );
    if ( access.poll.uGetStatus() == uPoll::uAlwaysPoll ) access.poll.uSetPollFlag( access.fd );
    ufile.access();
} // uFileAccess::uFileAccess


uFileAccess::~uFileAccess() {
    ufile.unaccess();
    if ( access.poll.uGetStatus() == uPoll::uAlwaysPoll ) access.poll.uClearPollFlag( access.fd );
    if ( access.fd >= 3 ) {				// don't close the standard file descriptors
	int code;

	uThisCluster().uCloseFD( access.fd );
	for ( ;; ) {
	    code = ::close( access.fd );
	  if ( code != -1 ) break;
	  if ( errno != EINTR ) break;			// timer interrupt ?
	} // for
	if ( code == -1 ) {
	    if ( ! std::uncaught_exception() ) _Resume uFileAccess::CloseFailure( *this, "unable to terminate access to file" );
	} // if
    } // if
} // uFileAccess::~uFileAccess


off_t uFileAccess::lseek( off_t offset, int whence ) {
    off_t code;

    for ( ;; ) {
	code = ::lseek( access.fd, offset, whence );
      if ( code != -1 ) break;
      if ( errno != EINTR ) break;			// timer interrupt ?
    } // for
    if ( code == -1 ) {
        _Throw uFileAccess::SeekFailure( *this, offset, whence, "could not seek file" );
    } // if
    return code;
} // uFileAccess::lseek


int uFileAccess::fsync() {
    int code;

    for ( ;; ) {
	code = ::fsync( access.fd );
      if ( code != -1 ) break;
      if ( errno != EINTR ) break;			// timer interrupt ?
    } // for
    if ( code == -1 ) {
        _Throw uFileAccess::SyncFailure( *this, "could not fsync file" );
    } // if
    return code;
} // uFileAccess::fsync


//######################### uFile (cont) #########################


uFile::Failure::Failure( const uFile &f, const char *const msg ) : uIOFailure( msg ), f( f ) {
    // file name is copied because its storage can be freed and scrubbed before handler starts
    uEHM::uStrncpy( uName, f.getName(), uEHMMaxName );
} // uFile::Failure::Failure

uFile::Failure::~Failure() {}

const uFile &uFile::Failure::file() const { return f; }

const char *uFile::Failure::getName() const { return uName; }

void uFile::Failure::defaultTerminate() const {
    uAbort( "(uFile &)0x%p, %.256s \"%.256s\".", &file(), message(), getName() );
} // uFile::Failure::defaultTerminate


uFile::TerminateFailure::TerminateFailure( const uFile &f, const int accessCnt, const char *const msg ) :
	uFile::Failure( f, msg ), accessCnt( accessCnt ) {}

void uFile::TerminateFailure::defaultTerminate() const {
    uAbort( "(uFile &)0x%p.~uFile(), %.256s, %d accessor(s) outstanding.", &file(), message(), accessCnt );
} // uFile::TerminateFailure::defaultTerminate


uFile::StatusFailure::StatusFailure( const uFile &f, const struct stat &buf, const char *const msg ) : uFile::Failure( f, msg ), buf( buf ) {}

void uFile::StatusFailure::defaultTerminate() const {
    uAbort( "(uFile &)0x%p.status( buf:0x%p ), %.256s \"%.256s\".\nError(%d) : %s.",
	    &file(), &buf, message(), getName(), errNo(), strerror( errNo() ) );
} // uFile::StatusFailure::defaultTerminate


//######################### uFileAccess (cont) #########################


uFileAccess::Failure::Failure( const uFileAccess &fa, const char *const msg ) : uFile::Failure( fa.ufile, msg ), fa( fa ) {
    fd = fa.access.fd;
} // uFileAccess::Failure::Failure

const uFileAccess &uFileAccess::Failure::fileAccess() const {
    return fa;
} // uFileAccess::Failure::fileAccess

int uFileAccess::Failure::fileDescriptor() const {
    return fd;
} // uFileAccess::Failure::fileDescriptor

void uFileAccess::Failure::defaultTerminate() const {
    uAbort( "(uFileAccess &)0x%p( file:0x%p ), %.256s file \"%.256s\".",
	    &fileAccess(), &file(), message(), getName() );
} // uFileAccess::Failure::defaultTerminate


uFileAccess::OpenFailure::OpenFailure( uFileAccess &fa, int flags, int mode, const char *const msg ) :
	uFileAccess::Failure( fa, msg ), flags( flags ), mode( mode ) {}

void uFileAccess::OpenFailure::defaultTerminate() const {
    uAbort( "(uFileAccess &)0x%p.uFileAccess( file:0x%p, flags:0x%x, mode:0x%x ), %.256s \"%.256s\".\nError(%d) : %s.",
	    &fileAccess(), &file(), flags, mode, message(), getName(), errNo(), strerror( errNo() ) );
} // uFile::OpenFailure::defaultTerminate


uFileAccess::CloseFailure::CloseFailure( uFileAccess &fa, const char *const msg ) : uFileAccess::Failure( fa, msg ) {}

void uFileAccess::CloseFailure::defaultTerminate() const {
    uAbort( "(uFileAccess &)0x%p.~uFileAccess(), %.256s \"%.256s\".\nError(%d) : %s.",
	    &fileAccess(), message(), getName(), errNo(), strerror( errNo() ) );
} // uFile::CloseFailure::defaultTerminate


uFileAccess::SeekFailure::SeekFailure( const uFileAccess &fa, const off_t offset, const int whence, const char *const msg ) :
	uFileAccess::Failure( fa, msg ), offset( offset ), whence( whence ) {}

void uFileAccess::SeekFailure::defaultTerminate() const {
    uAbort( "(uFile &)0x%p.lseek( offset:%ld, whence:%d ), %.256s file \"%.256s\".\nError(%d) : %s.",
	    &file(), (long int)offset, whence, message(), getName(), errNo(), strerror( errNo() ) );
} // uFileAccess::SeekFailure::defaultTerminate


uFileAccess::SyncFailure::SyncFailure( const uFileAccess &fa, const char *const msg ) : uFileAccess::Failure( fa, msg ) {}

void uFileAccess::SyncFailure::defaultTerminate() const {
    uAbort( "(uFileAccess &)0x%p.fsync(), %.256s \"%.256s\".\nError(%d) : %s.",
	    &file(), message(), getName(), errNo(), strerror( errNo() ) );
} // uFileAccess::SyncFailure::defaultTerminate


void uFileAccess::WriteFailure::defaultResume() const {
    if ( errNo() != EIO ) {
	_Throw *this;
    } // if
} // uFileAccess::WriteFailure::defaultResume


uFileAccess::ReadFailure::ReadFailure( const uFileAccess &fa, const char *buf, const int len, const uDuration *timeout, const char *const msg ) :
	uFileAccess::Failure( fa, msg ), buf( buf ), len( len ), timeout( timeout ) {}

void uFileAccess::ReadFailure::defaultTerminate() const {
    uAbort( "(uFileAccess &)0x%p.read( buf:0x%p, len:%d, timeout:0x%p ) : %.256s for file descriptor %d.\nError(%d) : %s.",
	    &fileAccess(), buf, len, timeout, message(), fileDescriptor(), errNo(), strerror( errNo() ) );
} // uFileAccess::ReadFailure::defaultTerminate

uFileAccess::ReadTimeout::ReadTimeout( const uFileAccess &fa, const char *buf, const int len, const uDuration *timeout, const char *const msg ) :
	uFileAccess::ReadFailure( fa, buf, len, timeout, msg ) {}


uFileAccess::WriteFailure::WriteFailure( const uFileAccess &fa, const char *buf, const int len, const uDuration *timeout, const char *const msg ) :
	uFileAccess::Failure( fa, msg ), buf( buf ), len( len ), timeout( timeout ) {}

void uFileAccess::WriteFailure::defaultTerminate() const {
    uAbort( "(uFileAccess &)0x%p.write( buf:0x%p, len:%d, timeout:0x%p ) : %.256s for file descriptor %d.\nError(%d) : %s.",
	    &fileAccess(), buf, len, timeout, message(), fileDescriptor(), errNo(), strerror( errNo() ) );
} // uFileAccess::WriteFailure::defaultTerminate

uFileAccess::WriteTimeout::WriteTimeout( const uFileAccess &fa, const char *buf, const int len, const uDuration *timeout, const char *const msg ) :
	uFileAccess::WriteFailure( fa, buf, len, timeout, msg ) {}


// Local Variables: //
// compile-command: "gmake install" //
// End: //
